# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

DEFAULT_AUTH_PATH = "/api/rest/v1/auth"

DEFAULT_JSON_API_PATH = "/api/json"
DEFAULT_JSON_API_VERSION = "v1"

DEFAULT_DEV_OPS_URL = "https://api.astra.datastax.com"
DEFAULT_DEV_OPS_API_VERSION = "v2"
DEFAULT_DEV_OPS_AUTH_HEADER = "Authorization"

DEFAULT_TIMEOUT = 30000
DEFAULT_AUTH_HEADER = "Token"
DEFAULT_KEYSPACE_NAME = "default_keyspace"
DEFAULT_REGION = "us-east1"

DEFAULT_INSERT_NUM_DOCUMENTS = 50

# Some of these are repeated by hand from idiomatic, tolerable duplication
# as long as `core` is in place:
#   (the case must match, so that secrets are effectively masked)
DEFAULT_REDACTED_HEADERS = {
    DEFAULT_DEV_OPS_AUTH_HEADER,
    DEFAULT_AUTH_HEADER,
    "X-Embedding-Api-Key",
    "X-Embedding-Access-Id",
    "X-Embedding-Secret-Id",
}
